#ifdef _MSC_VER
#pragma warning(disable:4786)
#endif

#include <map>
#include <omnetpp.h>
#include <vector>
#include <list>
#include "Packet_m.h"

using namespace omnetpp;


class Routing : public cSimpleModule
{
  private:
    int myAddress; //direccion del router propio
    int NumeroNodos;//tiempos de borrado de las tablas
    cPar *timeStore;
    cPar *timetolive;
    cPar *cleanretry;
    cMessage *cleanstoretime;
    cMessage *cleanretrytime;
    cMessage *cleanTemporal;
    simsignal_t arrivalSignal;
    int criterio;


  public:
    Routing();
    virtual ~Routing();

  protected:
    int retry;
    int numPorts;
    int inputGateBaseId;
    int outputGateBaseId;
    unsigned int rreqId = 0;
    unsigned int rreqCount = 0;

    /*unsigned int ttlIncrement = 2;
    unsigned int ttlThreshold = 7;
    unsigned int netDiameter = 35;
    unsigned int ttlStart = 2;
    unsigned int timeoutBuffer = 2;
    */
    //unsigned int rreqRetries = 2;
    //simtime_t nodeTraversalTime;


    virtual void initialize() override;
    virtual void handleMessage(cMessage *msg) override;
    virtual void forwardMessage(cMessage *msg);
    virtual void handlepack(cMessage *msg);
    virtual void handleRREQ(cMessage *msg);
    virtual void handleRREP(cMessage *msg);
    virtual void handleRRER(cMessage *msg);
    virtual void handleWaitForRREP(WaitForRREP *rrepTimer);
    virtual void handleWaitForRREQ(WaitForRREQ *rreqTimer);
    LOADngRREQ *createRREQ(int destAddr);
    LOADngRREP *createRREP(LOADngRREQ *rreq);
    virtual void refreshDisplay() const override;

    //tablas varias
    std::map<int, WaitForRREP *> waitForRREPTimers; //mapa con los temporizadores con el destino y su mensaje
    std::map<int, unsigned int> addressToRreqRetries; //mapa con los reintentos
    std::vector< std::vector<int> > direcciones; //tabla de enrutamiento //{origen, destino,nodo(donde se guarda la entrada),destino ,puerto ,saltos, enlacesdebiles}
    std::vector< std::vector<int> > temporal; //tabla temporal de los rreq //{origen,destino,nodo(donde se guarda la entrada), nodo de donde viene, puerto, saltos, enlacesdebiles,rreqID}
};
Define_Module(Routing);
Routing::Routing()
{
    cleanstoretime = nullptr;
    cleanretrytime =nullptr;
    cleanTemporal =nullptr;
}
Routing::~Routing()
{
    cancelAndDelete(cleanstoretime);
    cancelAndDelete(cleanretrytime);
    cancelAndDelete(cleanTemporal);
}
void Routing::initialize()

{
    arrivalSignal = registerSignal("arrival");
    timeStore = &par("timeStore");
    timetolive = &par("timetolive");
    retry = par("retry");
    cleanretry = &par("cleanretry");
    numPorts = gateSize("out");
    inputGateBaseId = gateBaseId("in");
    outputGateBaseId = gateBaseId("out");
    NumeroNodos = getAncestorPar("NumeroNodos"); //miro cuantos nodos tiene mi sistema
    direcciones.clear();
    temporal.clear();
    direcciones.push_back({0,0,0,0});
    temporal.push_back({0,0,0,0,0,0});
    myAddress = getParentModule()->par("address");
    cleanstoretime = new cMessage("cleanstoretime");
    scheduleAt(simTime(), cleanstoretime);
    cleanretrytime = new cMessage("cleanretrytime");
    scheduleAt(simTime(), cleanretrytime);
    //cleanTemporal = new cMessage("cleanTemporal");
    //scheduleAt(simTime() + 10.0, cleanTemporal);
    for (int i=0;i<100;i++)
    {
        addressToRreqRetries[i]=0;
    }
    criterio = par("criterio");

}
void Routing::handleMessage(cMessage *msg)
{
    EV<<"direcciones del nodo"<< myAddress <<endl;

    for (int i=1;i<direcciones.size();i++)
    {
        for(int j = 0; j<direcciones[i].size(); j++)
        {
            EV<<direcciones[i][j]<<",";
        }
        EV<<endl;
    }


    EV<<"Tabla ida"<<endl;
    for (int i=1;i<temporal.size();i++)
    {
        for(int j = 0; j<temporal[i].size(); j++)
        {
            EV<<temporal[i][j];
        }
        EV<<endl;

    }
    myAddress = getParentModule()->par("address");
    if(msg->isSelfMessage()){
        //si llega un cleanstoretime limpio las tablas de direciones
        if (msg == cleanstoretime) {
            EV<<"Borradas las tablas rutas"<<endl;
                for (int i=0;i<direcciones.size();i++){
                    if(direcciones[i][0]==myAddress){
                        direcciones.erase(direcciones.begin() + i);
                        //destinos.clear();
                    }
                }
        }
        //si llega un cleanretrytime borro la tabla de los reintentos
        if(msg == cleanretrytime)
        { EV<<"Borradas las tablas de reintentos"<<endl;
            for (int i=0;i<100;i++)
            {
                addressToRreqRetries[i]=0;
            }
        }
        else if(dynamic_cast<WaitForRREQ *>(msg)){
            handleWaitForRREQ((WaitForRREQ *)msg);

        }
        /*if (msg == cleanTemporal) {
            for (int i=0;i<temporal.size();i++){
                temporal[i].clear();
            }
            if (!cleanTemporal->isScheduled())
            {
            scheduleAt(simTime() + 10.0, cleanTemporal);}
        }*/


        //si llega un WaitForRREP
        else if(dynamic_cast<WaitForRREP *>(msg))
            handleWaitForRREP((WaitForRREP *)msg);
    }
    else {
    Packet *pk = check_and_cast<Packet *>(msg);
    //Ahora diferencio los mensajes si es pk, RREQ, RREP, RERR, y lo  trato segun toque
    switch (pk->getPacketType()) {
                case PACK:
                    if(pk->getHopCount()>NumeroNodos){
                        delete msg;
                        return;
                    }
                    handlepack(pk);
                    break;
                case RREQ:
                    if(pk->getHopCount()>NumeroNodos){
                        delete msg;
                        return;
                    }
                    handleRREQ(pk);
                    break;
                case RREP:
                    if(pk->getHopCount()>NumeroNodos){
                        delete msg;
                        return;
                    }
                    handleRREP(pk);
                    break;
                case RRER:
                    if(pk->getHopCount()>NumeroNodos){
                        delete msg;
                        return;
                    }
                    handleRRER(pk);
                    break;
    }
 return;
    }
}
void Routing::handlepack(cMessage *msg)
{
    myAddress = getParentModule()->par("address");
    Packet *pk = check_and_cast<Packet *>(msg);
    int destAddr = pk->getDestAddr();
    if (destAddr == myAddress){
    delete pk;
    }
       if (destAddr != myAddress){
           //si el destino es distinto al propio router busco si hay direccion para ello
           for (int i=0;i<direcciones.size();i++)
           {
               if (direcciones[i][0]==myAddress && direcciones[i][1]==pk->getDestAddr())
               {
                   cGate *ogate = gate(outputGateBaseId + direcciones[i][2]);
                   if (ogate->isConnected()){
                   send(pk->dup(), ogate);
                   EV<<"MENSAJE MANDADO POR RUTA"<<endl;}
                   delete pk;
                   return;
               }
           }
                   EV<<"No hay ruta en la tabla desde "<<myAddress<<" hasta "<<pk->getDestAddr()<<endl;
               //si no hay ruta creo el RREQ y comienzo la busqueda de la ruta y pongo su numero de reintentos a 0
               LOADngRREQ *rreq = createRREQ(destAddr);
               auto rreqCount=addressToRreqRetries.find(pk->getDestAddr());
               int reintentos=rreqCount->second;
               EV<<"destino "<<pk->getDestAddr()<<"reintentos "<<reintentos<<endl;
               if (reintentos >= retry) {
                   EV<< "Numero de reintentos de busqueda de ruta alcanzados. Cancelado el envio del RREQ con destino "<< destAddr << endl;
                   delete rreq;
                   if (!cleanretrytime->isScheduled())
                           {
                           scheduleAt(simTime() + cleanretry->doubleValue(), cleanretrytime);}
                   return;
               }
               delete pk;
               //inicio la temporaizacion de espera por el rrep
               auto rrepTimer=waitForRREPTimers.find(rreq->getDestAddr());
               WaitForRREP *rrepTimerMsg=nullptr;
               if(rrepTimer != waitForRREPTimers.end()){
                   rrepTimerMsg=rrepTimer->second;
                   rrepTimerMsg->setDestAddr(rreq->getDestAddr());
                   if (timetolive != 0) {
                        cancelEvent(rrepTimerMsg);
                    }
                }
                else {
                    rrepTimerMsg = new WaitForRREP();
                    waitForRREPTimers[rreq->getDestAddr()] = rrepTimerMsg;
                    rrepTimerMsg->setDestAddr(rreq->getDestAddr());
                }
               scheduleAt(simTime() + timetolive->doubleValue()/*ringTraversalTime*/, rrepTimerMsg);
               EV_INFO << "Comienzo busqueda de ruta a " << rreq->getDestAddr() << " con tiempo de respuesta de " << timetolive->doubleValue() << "en "<< simTime() <<  endl;
               forwardMessage(rreq);
               addressToRreqRetries[rreq->getDestAddr()]++;
               delete rreq;
               return;
       }
}
void Routing::handleRREQ(cMessage *msg)
{
    LOADngRREQ *rreq= check_and_cast<LOADngRREQ *>(msg);
    EV_INFO << "LOADng Route Request arrived with source addr: " << rreq->getOriginatorAddr()<< " destination addr: " << rreq->getDestAddr() << endl;
    //Cuando un nodo recibe un RREQ miro si ya tiene ruta con ese destino, si es asi comparo si es mejor o peor que la anterior
    //si es mejor reescribo la tabla, si es peor descarto el mensaje
    int destAddr = rreq->getDestAddr();
    int arrivalPort = rreq->getArrivalGate()->getIndex();
           if (destAddr != myAddress){
               //si llega de nuevo al origen, ha hecho un circulo, y lo borro
               if(rreq->getOriginatorAddr()==myAddress){
                   EV<<"el RREQ ha dado una vuelta en circulo, lo dejo de enviar"<<endl;
                   delete msg;
                   return;
               }
                   switch(criterio){
                   //criterio 0 cuenta numero de saltos y se queda con el de menos saltos
                   case 0:
                   {
                       for (int i=0;i<temporal.size();i++)
                         {if(temporal[i].empty()){
                         }
                         else{
                             //cuando llega a un nodo miro si ya ha pasado otro rreq con con menos saltos
                             if (temporal[i][0]==rreq->getOriginatorAddr() && temporal[i][1]==destAddr && temporal[i][2]==myAddress && temporal[i][7]==rreq->getRreqId())
                             {
                                 if(temporal[i][5]<=rreq->getHopCount())
                                 {//si encuentro uno con mas saltos y mismo origen lo e id lo dehecho
                                     EV<<"hay una ruta mas corta, dejo de reenviarlo"<<endl;
                                     delete msg;
                                     return;
                                 }
                                 else{
                                     //si ha llegado y tiene menos saltos que el anterior lo meto en la tabla
                                     for (int i=0;i<temporal.size();i++){
                                         if (temporal[i][0]==rreq->getOriginatorAddr() && temporal[i][1]==destAddr && temporal[i][2]==myAddress && temporal[i][7]==rreq->getRreqId()){
                                             temporal[i].clear();
                                             temporal[i]={rreq->getOriginatorAddr(),rreq->getDestAddr(),myAddress,rreq->getPrevio(),rreq->getArrivalGate()->getIndex(),rreq->getHopCount(),rreq->getWeak(),rreq->getRreqId()};
                                             rreq->setHopCount(rreq->getHopCount()+1);
                                             for(int i =0; i < numPorts; i++){
                                                 if (i!=arrivalPort){
                                                 cGate *ogate = gate(outputGateBaseId + i);
                                                 if (ogate->isConnected()){
                                                 rreq->setPrevio(myAddress);
                                                 send(rreq->dup(),ogate);}}}
                                                 delete msg;
                                                 return;}}
                                             temporal.push_back({rreq->getOriginatorAddr(),rreq->getDestAddr(),myAddress,rreq->getPrevio(),rreq->getArrivalGate()->getIndex(),rreq->getHopCount(),rreq->getWeak(),rreq->getRreqId()});
                                             rreq->setHopCount(rreq->getHopCount()+1);
                                                 for(int i =0; i < numPorts; i++){
                                                     if (i!=arrivalPort){
                                                     cGate *ogate = gate(outputGateBaseId + i);
                                                     if (ogate->isConnected()){
                                                     rreq->setPrevio(myAddress);
                                                     send(rreq->dup(),ogate);}}}
                                                     delete msg;
                                                     return;
                                     }
                             }
                            }}
                   }
                   case 1:{//camino con menor delay
                       for (int i=0;i<temporal.size();i++)
                         {if(temporal[i].empty()){
                         }
                         else{
                             //cuando llega a un nodo miro si ya ha pasado otro rreq
                             if (temporal[i][0]==rreq->getOriginatorAddr() && temporal[i][1]==destAddr && temporal[i][2]==myAddress && temporal[i][7]==rreq->getRreqId())
                             {//si encuentro uno con el mismo origen y destino lo deshecho y me quedo con el que ha llegado antes
                                     EV<<"hay una ruta mas rapida, dejo de reenviarlo"<<endl;
                                     delete msg;
                                     return;
                             }
                         }
                         }

                   }
                   case 2://camino con menores enlaces debiles
                       {
                        for (int i=0;i<temporal.size();i++)
                        {if(temporal[i].empty()){
                        }
                         else{
                              //cuando llega a un nodo miro si ya ha pasado otro rreq con menos enlaces debiles
                              if (temporal[i][0]==rreq->getOriginatorAddr() && temporal[i][1]==destAddr && temporal[i][2]==myAddress && temporal[i][7]==rreq->getRreqId())
                               {
                               if(temporal[i][6]<=rreq->getWeak())
                                 {//si encuentro uno con menos saltos y mismo origen lo deshecho
                                    EV<<"hay una ruta con menos enlaces debiles, dejo de reenviarlo"<<endl;
                                    delete msg;
                                    return;
                                    }
                                    else{
                                    //si ha llegado y tiene menos enlaces debiles que el anterior lo meto en la tabla
                                    for (int i=0;i<temporal.size();i++){
                                      if (temporal[i][0]==rreq->getOriginatorAddr() && temporal[i][1]==destAddr && temporal[i][2]==myAddress && temporal[i][7]==rreq->getRreqId()){
                                          temporal[i].clear();
                                          temporal[i]={rreq->getOriginatorAddr(),rreq->getDestAddr(),myAddress,rreq->getPrevio(),rreq->getArrivalGate()->getIndex(),rreq->getHopCount(),rreq->getWeak(),rreq->getRreqId()};
                                          rreq->setHopCount(rreq->getHopCount()+1);
                                          for(int i =0; i < numPorts; i++){
                                          if (i!=arrivalPort){
                                          cGate *ogate = gate(outputGateBaseId + i);
                                          if (ogate->isConnected()){
                                          rreq->setPrevio(myAddress);
                                          send(rreq->dup(),ogate);}}}
                                          delete msg;
                                          return;}}
                                        temporal.push_back({rreq->getOriginatorAddr(),rreq->getDestAddr(),myAddress,rreq->getPrevio(),rreq->getArrivalGate()->getIndex(),rreq->getHopCount(),rreq->getWeak(),rreq->getRreqId()});
                                        rreq->setHopCount(rreq->getHopCount()+1);
                                        for(int i =0; i < numPorts; i++){
                                        if (i!=arrivalPort){
                                         cGate *ogate = gate(outputGateBaseId + i);
                                         if (ogate->isConnected()){
                                         rreq->setPrevio(myAddress);
                                         send(rreq->dup(),ogate);}}}
                                         delete msg;
                                         return;
                                                        }
                                                }
                                               }}
                                      }


                   }
                            //si no hay ruta lo mando por el resto de puertos y actualizo la tabla temporal
                            for (int i=0;i<temporal.size();i++){
                                    if(temporal[i].empty()){
                                    temporal[i]={rreq->getOriginatorAddr(),rreq->getDestAddr(),myAddress,rreq->getPrevio(),rreq->getArrivalGate()->getIndex(),rreq->getHopCount(),rreq->getWeak(),rreq->getRreqId()};
                                    rreq->setHopCount(rreq->getHopCount()+1);
                                    for(int i =0; i < numPorts; i++){
                                        if (i!=arrivalPort){
                                        cGate *ogate = gate(outputGateBaseId + i);
                                        if (ogate->isConnected()){
                                        rreq->setPrevio(myAddress);
                                        send(rreq->dup(),ogate);}}}
                                    delete msg;
                                    return;}
                                    else{
                                    temporal.push_back({rreq->getOriginatorAddr(),rreq->getDestAddr(),myAddress,rreq->getPrevio(),rreq->getArrivalGate()->getIndex(),rreq->getHopCount(),rreq->getWeak(),rreq->getRreqId()});
                                    rreq->setHopCount(rreq->getHopCount()+1);
                                    for(int i =0; i < numPorts; i++){
                                        if (i!=arrivalPort){
                                        cGate *ogate = gate(outputGateBaseId + i);
                                        if (ogate->isConnected()){
                                        rreq->setPrevio(myAddress);
                                        send(rreq->dup(),ogate);}
                                        }
                                        }
                                    delete msg;
                                    return;}
                            }

           }
           else{//si llego al destino creo el rrep y lo mando por donde vino
               //recorro temporal y si no hay entrada con el origen y destino, es el primero en llegar e inicio la espera
               int encuentro=0;
               switch(criterio){
               case 0:{
               for (int i=0;i<temporal.size();i++){
                   if(temporal[i][0]==rreq->getOriginatorAddr() && temporal[i][1]==destAddr && temporal[i][7]==rreq->getRreqId()){
                       encuentro++;
                       if(temporal[i][5]<=rreq->getHopCount())
                       {//si encuentro uno con menos saltos y mismo origen lo deshecho
                           delete msg;
                           return;
                       }
                       else{
                           //si ha llegado y tiene menos saltos que el anterior lo meto en la tabla y reemplazo el anterior
                           for (int i=0;i<temporal.size();i++){
                               if (temporal[i][0]==rreq->getOriginatorAddr() && temporal[i][1]==destAddr && temporal[i][2]==myAddress && temporal[i][7]==rreq->getRreqId()){
                                   temporal[i].clear();
                                   temporal[i]={rreq->getOriginatorAddr(),rreq->getDestAddr(),myAddress,rreq->getPrevio(),rreq->getArrivalGate()->getIndex(),rreq->getHopCount(), rreq->getRreqId()};
                                   rreq->setHopCount(rreq->getHopCount()+1);
                                   LOADngRREP *rrep = createRREP(rreq);
                                   rrep->setPrevio(myAddress);
                                   cGate *ogate =gate(outputGateBaseId+arrivalPort);
                                   if (ogate->isConnected()){
                                   send(rrep,ogate);}
                                   delete rreq;
                                   return;
                   }}}}
               }
               }
               case 1:{//si llego al destino y ya hay entrada deshecho el mensaje
               for (int i=0;i<temporal.size();i++){
                    if(temporal[i][0]==rreq->getOriginatorAddr() && temporal[i][1]==destAddr && temporal[i][7]==rreq->getRreqId()){
                      encuentro++;
                      delete msg;
                      return;}}
               }
               case 2:{
                   for (int i=0;i<temporal.size();i++){
                       if(temporal[i][0]==rreq->getOriginatorAddr() && temporal[i][1]==destAddr && temporal[i][7]==rreq->getRreqId()){
                           encuentro++;
                           if(temporal[i][6]<=rreq->getWeak())
                           {//si encuentro uno con mas enlaces debiles y mismo origen lo deshecho
                               delete msg;
                               return;
                           }
                           else{
                               //si ha llegado y tiene menos enlaces debiles que el anterior lo meto en la tabla y reemplazo el anterior
                               for (int i=0;i<temporal.size();i++){
                                   if (temporal[i][0]==rreq->getOriginatorAddr() && temporal[i][1]==destAddr && temporal[i][2]==myAddress && temporal[i][7]==rreq->getRreqId()){
                                       temporal[i].clear();
                                       temporal[i]={rreq->getOriginatorAddr(),rreq->getDestAddr(),myAddress,rreq->getPrevio(),rreq->getArrivalGate()->getIndex(),rreq->getHopCount(),rreq->getWeak(),rreq->getRreqId()};
                                       rreq->setHopCount(rreq->getHopCount()+1);
                                       LOADngRREP *rrep = createRREP(rreq);
                                       rrep->setPrevio(myAddress);
                                       cGate *ogate =gate(outputGateBaseId+arrivalPort);
                                       if (ogate->isConnected()){
                                       send(rrep,ogate);}
                                       delete rreq;
                                       return;
                       }}}}
               }
               }
               }
               if (encuentro==0){
                   //si no hay entrada que coincida en la tabla significa que es la primera y tengo que empezar la temporizacion
                   /*EV<<"inicio temporizacion de 5s"<<endl;
                   WaitForRREQ *rreqespera=nullptr;
                   rreqespera = new WaitForRREQ();
                   rreqespera->setOrigen(rreq->getOriginatorAddr());
                   rreqespera->setDestino(rreq->getDestAddr());
                   scheduleAt(simTime()+ 5.0, rreqespera);
                   EV<<rreqespera->getOrigen()<<rreqespera->getDestino()<<endl;*/
                   temporal.push_back({rreq->getOriginatorAddr(),rreq->getDestAddr(),myAddress,rreq->getPrevio(),rreq->getArrivalGate()->getIndex(),rreq->getHopCount(),rreq->getWeak(),rreq->getRreqId()});
                   int arrivalPort = rreq->getArrivalGate()->getIndex();
                   LOADngRREP *rrep = createRREP(rreq);
                   rrep->setPrevio(myAddress);
                   cGate *ogate =gate(outputGateBaseId+arrivalPort);
                   if (ogate->isConnected()){
                   send(rrep,ogate);}
                   delete rreq;
                   return;
               }
  }

}
void Routing::handleRREP(cMessage *msg)
{
    //criterio 0 cuenta numero de saltos y se queda con el de menos saltos
LOADngRREP *rrep=check_and_cast<LOADngRREP *>(msg);
int repetido=0;
switch(criterio){
case 0:{
for (int i=0;i<direcciones.size();i++)
    //si hay una ruta con el mismo origen y el mismo numero de saltos, esta repetido, la deshecho por la que llego antes
{if (direcciones[i][0]==myAddress && direcciones[i][1]==rrep->getOriginatorAddr() && direcciones[i][3]<=rrep->getHopCount())
{repetido++;
EV<<"repetido"<<endl;
    }
}
//para el caso de menor numero de saltos
for (int i=0;i<direcciones.size();i++){
if (direcciones[i][0]==myAddress && direcciones[i][1]==rrep->getOriginatorAddr() && direcciones[i][3]>rrep->getHopCount())
{//si hay un camino con menos saltos lo borro para meter el nuevo
    EV<<"borro direccion y e introduzco nueva"<<endl;
    swap(direcciones[i],direcciones.back());
    direcciones.pop_back();
    }}
//si no está repetido lo meto en la tabla
if (repetido==0){
        int hopcount=rrep->getHopCount();
        emit(arrivalSignal, hopcount);
        direcciones.push_back({myAddress,rrep->getOriginatorAddr(),rrep->getArrivalGate()->getIndex(),rrep->getHopCount(),rrep->getWeak()});
        EV<<"meto la ruta en la tabla"<<endl;
        //cuando meto una ruta en la tabla, tengo que poner a 0 el nuemero de reintentos de busqueda, porque ya la encontre
        addressToRreqRetries[rrep->getOriginatorAddr()]=0;
        //aquí tengo que empezar a contar el tiempo para borrarlo
        if (!cleanstoretime->isScheduled())
        {
        scheduleAt(simTime() + timeStore->doubleValue(), cleanstoretime);}
}
if (rrep->getDestAddr()==myAddress)
{//si llega al destino, busqueda finalizada, lo borro
    delete msg;
return;}
    rrep->setHopCount(rrep->getHopCount()+1);
    int minsaltos=1000;
    int pos=0;
    for (int k=0;k<temporal.size();k++)
                       {if (temporal[k].empty()){
                       }
                       else{
                           if (temporal[k][0]==rrep->getDestAddr() && temporal[k][1]==rrep->getOriginatorAddr() && temporal[k][2]==myAddress && temporal[k][7]==rrep->getRrepid()){
                               if(temporal[k][5]<minsaltos){
                               minsaltos=temporal[k][5];
                               pos=k;}}}}

                               int n = gateSize("out");
                               for (int i=0;i<=n-1;i++){
                               cGate *outGate=gate("out",i);
                               cGate *inGate=outGate->getNextGate();
                               cModule *cola =inGate->getOwnerModule();
                               cGate *salida=cola->gate("line$o",-1);
                               cGate *previos=salida->getNextGate();
                               cGate *entradas=previos->getNextGate();
                               cModule *nodos=entradas->getOwnerModule();
                               int param = nodos->par("address");
                               if(param==temporal[pos][3]){
                                   send(rrep,"out",i);
                                   return;}

                               }

    }
case 1:{// para el caso de menor delay
for (int i=0;i<direcciones.size();i++)
        //si hay una ruta con el mismo origen y el mismo detino, esta repetido, la deshecho por la que llego antes
    {if (direcciones[i][0]==myAddress && direcciones[i][1]==rrep->getOriginatorAddr())
    {repetido++;
    EV<<"repetido"<<endl;
        }
    }
//si no está repetido lo meto en la tabla
    if (repetido==0){
            int hopcount=rrep->getHopCount();
            emit(arrivalSignal, hopcount);
            direcciones.push_back({myAddress,rrep->getOriginatorAddr(),rrep->getArrivalGate()->getIndex(),rrep->getHopCount(),rrep->getWeak()});
            EV<<"meto la ruta en la tabla"<<endl;
            //cuando meto una ruta en la tabla, tengo que poner a 0 el nuemero de reintentos de busqueda, porque ya la encontre
            addressToRreqRetries[rrep->getOriginatorAddr()]=0;
            //aquí tengo que empezar a contar el tiempo para borrarlo
            if (!cleanstoretime->isScheduled())
            {
            scheduleAt(simTime() + timeStore->doubleValue(), cleanstoretime);}
    }
    if (rrep->getDestAddr()==myAddress)
    {//si llega al destino, busqueda finalizada, lo borro
        delete msg;
    return;}
        rrep->setHopCount(rrep->getHopCount()+1);
        for (int k=0;k<temporal.size();k++)
                           {if (temporal[k].empty()){
                           }
                           else{
                               if (temporal[k][0]==rrep->getDestAddr() && temporal[k][1]==rrep->getOriginatorAddr() && temporal[k][2]==myAddress && temporal[k][7]==rrep->getRrepid()){
                                   int n = gateSize("out");
                                   for (int i=0;i<=n-1;i++){
                                   cGate *outGate=gate("out",i);
                                   cGate *inGate=outGate->getNextGate();
                                   cModule *cola =inGate->getOwnerModule();
                                   cGate *salida=cola->gate("line$o",-1);
                                   cGate *previos=salida->getNextGate();
                                   cGate *entradas=previos->getNextGate();
                                   cModule *nodos=entradas->getOwnerModule();
                                   int param = nodos->par("address");
                                   if(param==temporal[k][3]){
                                       send(rrep,"out",i);
                                       return;}
                                   }}
}
}
}
case 2:{//caso menos enlaces debiles
    for (int i=0;i<direcciones.size();i++)
        //si hay una ruta con el mismo origen y el mismo numero de enlaces debiles, esta repetido, la deshecho por la que llego antes
    {if (direcciones[i][0]==myAddress && direcciones[i][1]==rrep->getOriginatorAddr() && direcciones[i][4]<=rrep->getWeak())
    {repetido++;
    EV<<"repetido"<<endl;
        }
    }
    //para el caso de menor numero de enlaces débiles
    for (int i=0;i<direcciones.size();i++){
    if (direcciones[i][0]==myAddress && direcciones[i][1]==rrep->getOriginatorAddr() && direcciones[i][4]>rrep->getWeak())
    {//si hay un camino con menos enlaces debiles lo borro para meter el nuevo
        EV<<"borro direccion y e introduzco nueva"<<endl;
        swap(direcciones[i],direcciones.back());
        direcciones.pop_back();
        }}
    //si no está repetido lo meto en la tabla
    if (repetido==0){
            int hopcount=rrep->getHopCount();
            emit(arrivalSignal, hopcount);
            direcciones.push_back({myAddress,rrep->getOriginatorAddr(),rrep->getArrivalGate()->getIndex(),rrep->getHopCount(),rrep->getWeak()});
            EV<<"meto la ruta en la tabla"<<endl;
            //cuando meto una ruta en la tabla, tengo que poner a 0 el numero de reintentos de busqueda, porque ya la encontre
            addressToRreqRetries[rrep->getOriginatorAddr()]=0;
            //aquí tengo que empezar a contar el tiempo para borrarlo
            if (!cleanstoretime->isScheduled())
            {
            scheduleAt(simTime() + timeStore->doubleValue(), cleanstoretime);}
    }
    if (rrep->getDestAddr()==myAddress)
    {//si llega al destino, busqueda finalizada, lo borro
        delete msg;
    return;}
        rrep->setHopCount(rrep->getHopCount()+1);
        int minweak=1000;
        int pos=0;
        for (int k=0;k<temporal.size();k++)
                           {if (temporal[k].empty()){
                           }
                           else{
                               if (temporal[k][0]==rrep->getDestAddr() && temporal[k][1]==rrep->getOriginatorAddr() && temporal[k][2]==myAddress && temporal[k][7]==rrep->getRrepid()){
                                   if(temporal[k][6]<minweak){
                                   minweak=temporal[k][6];
                                   pos=k;}}}}

                                   int n = gateSize("out");
                                   for (int i=0;i<=n-1;i++){
                                   cGate *outGate=gate("out",i);
                                   cGate *inGate=outGate->getNextGate();
                                   cModule *cola =inGate->getOwnerModule();
                                   cGate *salida=cola->gate("line$o",-1);
                                   cGate *previos=salida->getNextGate();
                                   cGate *entradas=previos->getNextGate();
                                   cModule *nodos=entradas->getOwnerModule();
                                   int param = nodos->par("address");
                                   if(param==temporal[pos][3]){
                                       send(rrep,"out",i);
                                       return;}
                                   }
}
}
}
void Routing::handleRRER(cMessage *msg)
{
    //el rrer vuelve en todas direcciones y borra las tablas de cada router que tenga su direccion
    LOADngRRER *rrer= check_and_cast<LOADngRRER *>(msg);
    int arrivalPort = rrer->getArrivalGate()->getIndex();
    int encuentra=0;
    EV<<"puerto de llegada"<<myAddress<< rrer->getOriginatorAddr() <<arrivalPort<<endl;
    for (int i=0;i<direcciones.size();i++)
            {
                if(direcciones[i][0]==myAddress && direcciones[i][1]==rrer->getOriginatorAddr() && direcciones[i][2]==arrivalPort)
                {
                    swap(direcciones[i],direcciones.back());
                    direcciones.pop_back();
                    encuentra++;
                }
            }
    int total=temporal.size();
    for(int i=1;i<total;i++)
    {
        temporal.pop_back();
    }
    if (encuentra!=0){
    for(int i =0; i < numPorts; i++){
        if (i!=arrivalPort){
        cGate *ogate = gate(outputGateBaseId + i);
        if (ogate->isConnected()){
        send(rrer->dup(),ogate);
        }}
    }
    delete msg;
    }
    else{
        delete msg;
        return;
    }
    }

/*
    //el ensaje de error debe hacer el mismo camino de vuelta que el rrep que introdujo la tabla en los routers
    LOADngRRER *rrer=check_and_cast<LOADngRRER *>(msg);
    //si es el propio nodo el desconectado borro la entrada en su tabla
    if(rrer->getDestAddr()==myAddress)
    {
        for (int i=0;i<direcciones.size();i++)
        {
            if(direcciones[i][0]==myAddress && direcciones[i][1]==rrer->getOriginatorAddr())
            {
                swap(direcciones[i],direcciones.back());
                direcciones.pop_back();
                delete msg;
                return;
            }
        }
    }
    //si no es el propio nodo, hago el mismo camino que hizo el rrep y voy borrando las rutas
    else{
        for (int i=0;i<direcciones.size();i++)
        {
            if(direcciones[i][0]==myAddress && direcciones[i][1]==rrer->getOriginatorAddr())
            {
                swap(direcciones[i],direcciones.back());
                direcciones.pop_back();
            }
        }
    for (int i=0;i<temporal.size();i++)
                          {
                              if (temporal[i][2]==myAddress && temporal[i][3]==rrer->getDestAddr()){
                                  cGate *ogate = gate(outputGateBaseId + temporal[i][4]);
                                  if (ogate->isConnected()){
                              send(rrer,ogate);
                              return;}}
                          }*/

LOADngRREQ *Routing::createRREQ( int destAddr)
{
    LOADngRREQ *rreqPacket = new LOADngRREQ("LOADng-RREQ");
    rreqPacket->setPacketType(RREQ);
    rreqPacket->setOriginatorAddr(myAddress);
    rreqPacket->setPrevio(myAddress);
    rreqPacket->setDestAddr(destAddr);
    rreqId++;
    rreqPacket->setRreqId(rreqId);
    rreqPacket->setHopCount(0);
    rreqPacket->setWeak(0);
    return rreqPacket;
}
LOADngRREP *Routing::createRREP(LOADngRREQ *rreq)
{
    LOADngRREP *rrepPacket = new LOADngRREP("LOADng-RREP");
    rrepPacket->setPacketType(RREP);
    rrepPacket->setOriginatorAddr(myAddress);
    rrepPacket->setDestAddr(rreq->getOriginatorAddr());
    rrepPacket->setHopCount(1);
    rrepPacket->setWeak(rreq->getWeak());
    rrepPacket->setRrepid(rreq->getRreqId());
    return rrepPacket;
    }
void Routing::forwardMessage(cMessage *msg)
{
LOADngRREQ *rreq= check_and_cast<LOADngRREQ *>(msg);
rreq->setHopCount(rreq->getHopCount()+1);
for(int i =0; i < numPorts; i++)
{cGate *ogate = gate(outputGateBaseId + i);
if (ogate->isConnected()){
send(rreq->dup(),ogate);}}
return;
}
void Routing::handleWaitForRREP(WaitForRREP *rrepTimer)
{
    EV_INFO<<"LLEGA LA TEMPORIZACION DE LA ESPERA AL NODO"<<myAddress<<endl;

    for (int i=0;i<direcciones.size();i++){
        if (direcciones[i][1]==rrepTimer->getDestAddr()){
            EV<<"HAY RUTA ENCONTRADA EN EL TIEMPO DE ESPERA EN EL NODO "<<myAddress<< "CON DESTINO "<<rrepTimer->getDestAddr()<<endl;
            return;
        }
    }
        //else{
            EV<<"NO HAY RUTA ENCONTRADA EN EL TIEMPO DE ESPERA EN EL NODO "<<myAddress<< "CON DESTINO "<<rrepTimer->getDestAddr()<<"LUEGO VUELVO A MANDAR EL RREQ"<<endl;
            //vuelvo a iniciar la temporizacion a la espera de respuesta
                           auto rrepTimer2=waitForRREPTimers.find(rrepTimer->getDestAddr());
                           WaitForRREP *rrepTimerMsg=nullptr;
                           if(rrepTimer2 != waitForRREPTimers.end()){
                               rrepTimerMsg=rrepTimer2->second;
                               rrepTimerMsg->setDestAddr(rrepTimer->getDestAddr());
                               if (timetolive != 0) {
                                    cancelEvent(rrepTimerMsg);
                                }

                            }
                            else {
                                rrepTimerMsg = new WaitForRREP();
                                waitForRREPTimers[rrepTimer->getDestAddr()] = rrepTimerMsg;
                                rrepTimerMsg->setDestAddr(rrepTimer->getDestAddr());
                            }
                           scheduleAt(simTime() + timetolive->doubleValue(), rrepTimerMsg);
            int destAddr = rrepTimer->getDestAddr();
            ASSERT(addressToRreqRetries.find(destAddr) != addressToRreqRetries.end());

            if (addressToRreqRetries[destAddr] >= retry) {
                EV<< "Reintentos de busqueda para el destino " << destAddr << " es " << addressToRreqRetries[destAddr] << " dejo de enviar al RREQ" << endl;
                return;
            }
            else{
                LOADngRREQ *rreq = createRREQ(destAddr);
                //si no ha superado el numero de reintentos lo vuelvo a mandar
                addressToRreqRetries[destAddr]++;
                forwardMessage(rreq);
            }
            EV<< "Reintentos de busqueda para el destino " << destAddr << " es " << addressToRreqRetries[destAddr]<<endl;
        }
  //  }
    //EV_INFO << "No se ha obtenido respuesta durante el tiempo de espera" << endl;


//}
void Routing::handleWaitForRREQ(WaitForRREQ *rreqTimer){
    ///no hace nada por ahora, posible futura mejora
    /*
EV<<"llega la temporizacion con la info"<<rreqTimer->getOrigen()<<rreqTimer->getDestino()<<endl;
EV<<"temporal";
for (int i=0;i<temporal.size();i++)
{
    for(int j = 0; j<temporal[i].size(); j++)
    {
        EV<<temporal[i][j];
    }
    EV<<endl;
}
LOADngRREP *rrep=nullptr;
rrep = new LOADngRREP();
rrep->setPacketType(RREP);
rrep->setOriginatorAddr(myAddress);
rrep->setDestAddr(rreqTimer->getOrigen());
rrep->setHopCount(1);

for (int i=0;i<temporal.size();i++){
    if (temporal[i][0]==rrep->getDestAddr() && temporal[i][1]==rrep->getOriginatorAddr()){
        rrep->setPrevio(temporal[i][3]);
        cGate *ogate =gate(outputGateBaseId+temporal[i][4]);
        if (ogate->isConnected()){
        send(rrep,ogate);}
        return;}
}*/}

void Routing::refreshDisplay() const
{
    char *buf="";
    std::string str= std::string("Direcciones encontradas:\nOrigen|Destino|Puerto|Saltos|Debiles\n");
    for (int i=0;i<direcciones.size();i++)
    {if (direcciones[i][0]==0 && direcciones[i][1]==0)
    {}
    else{
        str+=std::to_string(direcciones[i][0])+"|"+std::to_string(direcciones[i][1])+"|"+std::to_string(direcciones[i][2])+"|"+std::to_string(direcciones[i][3])+"|"+std::to_string(direcciones[i][4])+"\n";
        buf =const_cast<char*>(str.c_str());}
    }
                getDisplayString().setTagArg("t",0,buf);

}





